from types import NoneType
from typing import List, Optional, Set, Tuple, Type

import pytest

from jmux.awaitable import (
    AwaitableValue,
    StreamableValues,
    UnderlyingGenericMixin,
)
from jmux.demux import JMux
from jmux.helpers import deconstruct_type, extract_types_from_generic_alias


@pytest.mark.parametrize(
    "TargetType,expected_tuple",
    [
        (int, {int}),
        (Optional[int], {int, NoneType}),
        (int | None, {int, NoneType}),
        (int | str, {str, int}),
        (int | str | NoneType, {str, int, NoneType}),
        (JMux, {JMux}),
        (JMux | None, {JMux, NoneType}),
        (JMux | NoneType, {JMux, NoneType}),
    ],
)
def test_extract_types(
    TargetType: Type[UnderlyingGenericMixin], expected_tuple: Tuple[Type, Set[Type]]
):
    underlying_types = deconstruct_type(TargetType)

    assert underlying_types == expected_tuple


class NestedObject(JMux):
    key: AwaitableValue[str]


# fmt: off
@pytest.mark.parametrize(
    "TargetType,expected_tuple",
    [
        (int, ({int}, set())),
        (str, ({str}, set())),
        (Optional[int], ({int, NoneType}, set())),
        (int | None, ({int, NoneType}, set())),
        (List[int], ({list}, {int})),
        (List[int | None], ({list}, {int, NoneType})),
        (AwaitableValue[int], ({AwaitableValue}, {int})),
        (AwaitableValue[float], ({AwaitableValue}, {float})),
        (AwaitableValue[str], ({AwaitableValue}, {str})),
        (AwaitableValue[bool], ({AwaitableValue}, {bool})),
        (AwaitableValue[NestedObject], ({AwaitableValue}, {NestedObject})),
        (AwaitableValue[int | None], ({AwaitableValue}, {int, NoneType})),
        (AwaitableValue[float | None], ({AwaitableValue}, {float, NoneType})),
        (AwaitableValue[str | None], ({AwaitableValue}, {str, NoneType})),
        (AwaitableValue[bool | None], ({AwaitableValue}, {bool, NoneType})),
        (AwaitableValue[NestedObject | None], ({AwaitableValue}, {NestedObject, NoneType})),
        (StreamableValues[int], ({StreamableValues}, {int})),
        (StreamableValues[float], ({StreamableValues}, {float})),
        (StreamableValues[str], ({StreamableValues}, {str})),
        (StreamableValues[bool], ({StreamableValues}, {bool})),
        (StreamableValues[NestedObject], ({StreamableValues}, {NestedObject})),
    ],
)
# fmt: on
def test_extract_types_from_generic_alias(
    TargetType: Type[UnderlyingGenericMixin], expected_tuple: Tuple[Type, Set[Type]]
):
    underlying_types = extract_types_from_generic_alias(TargetType)

    assert underlying_types == expected_tuple



from abc import ABC
from enum import Enum
from types import NoneType
from typing import (
    Optional,
    Set,
    Type,
    get_args,
    get_origin,
    get_type_hints,
)

from pydantic import BaseModel

from jmux.awaitable import AwaitableValue, IAsyncSink, SinkType, StreamableValues
from jmux.decoder import IDecoder, StringEscapeDecoder
from jmux.error import (
    EmptyKeyError,
    ForbiddenTypeHintsError,
    MissingAttributeError,
    NoCurrentSinkError,
    NotAllObjectPropertiesSetError,
    NothingEmittedError,
    ObjectAlreadyClosedError,
    ObjectMissmatchedError,
    ParsePrimitiveError,
    TypeEmitError,
    UnexpectedAttributeTypeError,
    UnexpectedCharacterError,
    UnexpectedStateError,
)
from jmux.helpers import (
    extract_types_from_generic_alias,
    get_main_type,
    is_json_whitespace,
)
from jmux.pda import PushDownAutomata
from jmux.types import (
    ARRAY_CLOSE,
    ARRAY_OPEN,
    BOOLEAN_ALLOWED,
    BOOLEAN_OPEN,
    COLON,
    COMMA,
    FLOAT_ALLOWED,
    INTERGER_ALLOWED,
    JSON_FALSE,
    JSON_NULL,
    JSON_TRUE,
    NULL_ALLOWED,
    NULL_OPEN,
    NUMBER_OPEN,
    OBJECT_CLOSE,
    OBJECT_OPEN,
    PRIMITIVE_STATES,
    QUOTE,
)
from jmux.types import Mode as M
from jmux.types import State as S

type Primitive = int | float | str | bool | None
type Emittable = Primitive | "JMux" | Enum


class Sink[T: Emittable]:
    def __init__(self, delegate: "JMux"):
        self._current_key: Optional[str] = None
        self._current_sink: Optional[IAsyncSink[T]] = None
        self._delegate: "JMux" = delegate

    @property
    def current_sink_type(self) -> SinkType:
        if self._current_sink is None:
            raise NoCurrentSinkError()
        return self._current_sink.get_sink_type()

    @property
    def current_underlying_generics(self) -> Set[Type[T]]:
        if self._current_sink is None:
            raise NoCurrentSinkError()
        return self._current_sink.get_underlying_generics()

    @property
    def current_underlying_main_generic(self) -> Type[T]:
        if self._current_sink is None:
            raise NoCurrentSinkError()
        return self._current_sink.get_underlying_main_generic()

    def set_current(self, attr_name: str) -> None:
        if not hasattr(self._delegate, attr_name):
            raise MissingAttributeError(
                object_name=self._delegate.__class__.__name__,
                attribute=attr_name,
            )
        sink = getattr(self._delegate, attr_name)
        if not isinstance(sink, IAsyncSink):
            raise UnexpectedAttributeTypeError(
                attribute=attr_name,
                object_name=type(sink).__name__,
                expected_type="IAsyncSink",
            )
        self._current_key = attr_name
        self._current_sink = sink

    async def emit(self, val: T) -> None:
        if self._current_sink is None:
            raise NoCurrentSinkError()
        generics = self._current_sink.get_underlying_generics()
        if not any(
            isinstance(val, underlying_generic) for underlying_generic in generics
        ):
            raise TypeEmitError(
                expected_type=f"{generics}",
                actual_type=f"{type(val).__name__}",
            )
        await self._current_sink.put(val)

    async def close(self) -> None:
        if self._current_sink is None:
            raise NoCurrentSinkError()
        await self._current_sink.close()

    async def ensure_closed(self) -> None:
        if self._current_sink is None:
            raise NoCurrentSinkError()
        await self._current_sink.ensure_closed()

    async def create_and_emit_nested(self) -> None:
        if self._current_sink is None:
            raise NoCurrentSinkError()
        NestedJmux = self._current_sink.get_underlying_main_generic()
        if not issubclass(NestedJmux, JMux):
            raise TypeEmitError(
                expected_type="JMux",
                actual_type=f"{NestedJmux.__name__}",
            )
        nested = NestedJmux()
        await self.emit(nested)

    async def forward_char(self, ch: str) -> None:
        if self._current_sink is None:
            raise NoCurrentSinkError()
        maybe_jmux = self._current_sink.get_current()
        if not isinstance(maybe_jmux, JMux):
            raise TypeEmitError(
                expected_type="JMux",
                actual_type=f"{type(maybe_jmux).__name__}",
            )
        await maybe_jmux.feed_char(ch)


class JMux(ABC):
    """
    JMux is an abstract base class for creating JSON demultiplexers.
    It parses a JSON stream and demultiplexes it into different sinks,
    which can be either AwaitableValue or StreamableValues.
    """

    def __init__(self):
        self._instantiate_attributes()
        self._pda: PushDownAutomata[M, S] = PushDownAutomata[M, S](S.START)
        self._decoder: IDecoder = StringEscapeDecoder()
        self._sink = Sink[Emittable](self)

    def _instantiate_attributes(self) -> None:
        type_hints = get_type_hints(self.__class__)
        for attr_name, type_alias in type_hints.items():
            TargetType = get_origin(type_alias)
            type_alias_args = get_args(type_alias)
            if len(type_alias_args) != 1:
                raise TypeError(f"Generic type {type_alias} must be fully specified")
            TargetGenericType = type_alias_args[0]
            target_instance = TargetType[TargetGenericType]()
            if not issubclass(TargetType, IAsyncSink):
                raise TypeError(
                    f"Attribute '{attr_name}' must conform to protocol IAsyncSink, "
                    f"got {TargetType}."
                )
            setattr(self, attr_name, target_instance)

    @classmethod
    def assert_conforms_to(cls, pydantic_model: Type[BaseModel]) -> None:
        """
        Asserts that the JMux class conforms to a given Pydantic model.

        Args:
            pydantic_model: The Pydantic model to compare against.

        Raises:
            MissingAttributeError: If an attribute is missing in the Pydantic model.
            ForbiddenTypeHintsError: If a type hint is not allowed.
            ObjectMissmatchedError: If the JMux class does not match the Pydantic model.
        """
        for attr_name, type_alias in get_type_hints(cls).items():
            jmux_main_type_set, jmux_subtype_set = extract_types_from_generic_alias(
                type_alias
            )

            MaybePydanticType = get_type_hints(pydantic_model).get(attr_name, None)
            if MaybePydanticType is None:
                if NoneType in jmux_subtype_set:
                    continue
                else:
                    raise MissingAttributeError(
                        object_name=pydantic_model.__name__,
                        attribute=attr_name,
                    )

            pydantic_main_type_set, pydantic_subtype_set = (
                extract_types_from_generic_alias(MaybePydanticType)
            )
            cls._assert_only_allowed_types(
                jmux_main_type_set,
                jmux_subtype_set,
                pydantic_main_type_set,
                pydantic_subtype_set,
            )
            if (
                pydantic_wrong := len(pydantic_main_type_set) != 1
                and len(pydantic_subtype_set) > 0
            ) or len(jmux_main_type_set) != 1:
                wrong_obj = "pydantic" if pydantic_wrong else "JMux"
                wrong_set = (
                    jmux_main_type_set if pydantic_wrong else pydantic_main_type_set
                )
                raise ForbiddenTypeHintsError(
                    message=(f"Forbidden typing received on {wrong_obj}: {wrong_set}"),
                )

            if StreamableValues in jmux_main_type_set:
                cls._assert_is_allowed_streamable_values(
                    jmux_subtype_set,
                    pydantic_subtype_set,
                    pydantic_main_type_set,
                    pydantic_model,
                    attr_name,
                )
            elif AwaitableValue in jmux_main_type_set:
                cls._assert_is_allowed_awaitable_value(
                    jmux_subtype_set,
                    pydantic_subtype_set,
                    pydantic_main_type_set,
                    pydantic_model,
                    attr_name,
                )
            else:
                raise ObjectMissmatchedError(
                    jmux_model=cls.__name__,
                    pydantic_model=pydantic_model.__name__,
                    attribute=attr_name,
                    message="Unexpected main type on JMux",
                )

    @classmethod
    def _assert_is_allowed_streamable_values(
        cls,
        jmux_subtype_set: Set[Type],
        pydantic_subtype_set: Set[Type],
        pydantic_main_type_set: Set[Type],
        pydantic_model: Type[BaseModel],
        attr_name: str,
    ):
        if len(jmux_subtype_set) != 1:
            raise ForbiddenTypeHintsError(
                "StreamableValues must have exactly one underlying type, "
                f"got {jmux_subtype_set}."
            )

        if list in pydantic_main_type_set:
            NonNoneType = get_main_type(jmux_subtype_set)
            PydanticNonNoneType = get_main_type(pydantic_subtype_set)
            if issubclass(NonNoneType, JMux):
                NonNoneType.assert_conforms_to(PydanticNonNoneType)
                return
            if jmux_subtype_set != pydantic_subtype_set:
                raise ObjectMissmatchedError(
                    jmux_model=cls.__name__,
                    pydantic_model=pydantic_model.__name__,
                    attribute=attr_name,
                    message=(
                        "StreamableValues of type list with subtype "
                        f"{jmux_subtype_set} does not match pydantic model "
                        f"type: {pydantic_subtype_set}"
                    ),
                )
        elif str in pydantic_main_type_set:
            if jmux_subtype_set != pydantic_main_type_set:
                raise ObjectMissmatchedError(
                    jmux_model=cls.__name__,
                    pydantic_model=pydantic_model.__name__,
                    attribute=attr_name,
                    message=(
                        "StreamableValues of type string does not match pydantic "
                        f"model type: {pydantic_main_type_set}"
                    ),
                )
        else:
            raise ForbiddenTypeHintsError(
                message="StreamableValues must be initialized with a sequence type."
            )

    @classmethod
    def _assert_is_allowed_awaitable_value(
        cls,
        jmux_subtype_set: Set[Type],
        pydantic_subtype_set: Set[Type],
        pydantic_main_type_set: Set[Type],
        pydantic_model: Type[BaseModel],
        attr_name: str,
    ):
        if len(pydantic_subtype_set) > 0:
            raise ForbiddenTypeHintsError(
                message="Pydantic model cannot have subtype for AwaitableValue."
            )
        NonNoneType = get_main_type(jmux_subtype_set)
        PydanticNonNoneType = get_main_type(pydantic_main_type_set)
        if issubclass(NonNoneType, JMux):
            NonNoneType.assert_conforms_to(PydanticNonNoneType)
            return
        if not jmux_subtype_set == pydantic_main_type_set:
            raise ObjectMissmatchedError(
                jmux_model=cls.__name__,
                pydantic_model=pydantic_model.__name__,
                attribute=attr_name,
                message=(
                    f"AwaitableValue with type {jmux_subtype_set} does not match "
                    f"pydantic model type: {pydantic_main_type_set}"
                ),
            )

    @classmethod
    def _assert_only_allowed_types(
        cls,
        jmux_main_type_set: Set[Type],
        jmux_subtype_set: Set[Type],
        pydantic_main_type_set: Set[Type],
        pydantic_subtype_set: Set[Type],
    ) -> None:
        if not all(t in (AwaitableValue, StreamableValues) for t in jmux_main_type_set):
            raise ForbiddenTypeHintsError(
                message=(
                    "JMux must have either AwaitableValue or StreamableValues as "
                    f"main type, got {jmux_main_type_set}."
                )
            )

        if not any(
            issubclass(elem, t)
            for t in (int, float, str, bool, NoneType, JMux, Enum)
            for elem in jmux_subtype_set
        ):
            raise ForbiddenTypeHintsError(
                message=(
                    "JMux sub type must be one of the emittable types: "
                    f"{jmux_subtype_set}."
                )
            )

        if len(pydantic_subtype_set) > 0 and not any(
            issubclass(elem, t)
            for t in (int, float, str, bool, NoneType, BaseModel, Enum)
            for elem in pydantic_subtype_set
        ):
            raise ForbiddenTypeHintsError(
                message=(
                    "Pydantic sub type must be one of the primitive, enum or "
                    f"BaseModel, got: {pydantic_subtype_set}."
                )
            )

        if not any(
            issubclass(elem, t)
            for t in (int, float, str, bool, list, NoneType, BaseModel, Enum)
            for elem in pydantic_main_type_set
        ):
            raise ForbiddenTypeHintsError(
                message=(
                    "Pydantic main type must be one of the primitive, enum, list "
                    f"or BaseModel, got {pydantic_main_type_set}."
                )
            )

    async def feed_chunks(self, chunks: str) -> None:
        """
        Feeds a string of characters to the JMux parser.

        Args:
            chunks: The string of characters to feed.

        Raises:
            UnexpectedCharacterError: If an unexpected character is encountered.
            ObjectAlreadyClosedError: If the JMux object is already closed.
            UnexpectedStateError: If the parser is in an unexpected state.
            EmptyKeyError: If an empty key is encountered in a JSON object.
        """
        for ch in chunks:
            await self.feed_char(ch)

    async def feed_char(self, ch: str) -> None:
        """
        Feeds a character to the JMux parser.

        Args:
            ch: The character to feed.

        Raises:
            UnexpectedCharacterError: If an unexpected character is encountered.
            ObjectAlreadyClosedError: If the JMux object is already closed.
            UnexpectedStateError: If the parser is in an unexpected state.
            EmptyKeyError: If an empty key is encountered in a JSON object.
        """
        if len(ch) != 1:
            raise UnexpectedCharacterError(
                character=ch,
                pda_stack=self._pda.stack,
                pda_state=self._pda.state,
                message="Only single characters are allowed to be fed to JMux.",
            )
        match self._pda.top:
            # CONTEXT: Start
            case None:
                match self._pda.state:
                    case S.START:
                        if ch in OBJECT_OPEN:
                            self._pda.push(M.ROOT)
                            self._pda.set_state(S.EXPECT_KEY)
                        else:
                            raise UnexpectedCharacterError(
                                ch,
                                self._pda.stack,
                                self._pda.state,
                                "JSON must start with '{' character.",
                            )
                    case S.END:
                        raise ObjectAlreadyClosedError(
                            object_name=self.__class__.__name__,
                            message=(
                                "Cannot feed more characters to closed JMux "
                                f"object, got '{ch}'"
                            ),
                        )
                    case _:
                        raise UnexpectedStateError(
                            self._pda.stack,
                            self._pda.state,
                            message=(
                                "Only START and END states are allowed in the root "
                                "context."
                            ),
                        )

            # CONTEXT: Root
            case M.ROOT:
                match self._pda.state:
                    case S.EXPECT_KEY:
                        if is_json_whitespace(ch):
                            pass
                        elif ch == '"':
                            self._pda.set_state(S.PARSING_KEY)
                            self._decoder.reset()
                        else:
                            raise UnexpectedCharacterError(
                                ch,
                                self._pda.stack,
                                self._pda.state,
                                "Char needs to be '\"' or JSON whitespaces",
                            )

                    case S.PARSING_KEY:
                        if self._decoder.is_terminating_quote(ch):
                            buffer = self._decoder.buffer
                            if not buffer:
                                raise EmptyKeyError(
                                    "Empty key is not allowed in JSON objects."
                                )
                            self._sink.set_current(buffer)
                            self._decoder.reset()
                            self._pda.set_state(S.EXPECT_COLON)
                        else:
                            self._decoder.push(ch)

                    case S.EXPECT_COLON:
                        if is_json_whitespace(ch):
                            pass
                        elif ch in COLON:
                            self._pda.set_state(S.EXPECT_VALUE)
                        else:
                            raise UnexpectedCharacterError(
                                ch,
                                self._pda.stack,
                                self._pda.state,
                                "Char must be ':' or JSON whitespaces.",
                            )

                    case S.EXPECT_VALUE:
                        if is_json_whitespace(ch):
                            pass
                        elif res := await self._handle_common__expect_value(ch):
                            if (
                                self._sink.current_sink_type
                                is SinkType.STREAMABLE_VALUES
                                and res is not S.PARSING_STRING
                            ):
                                raise UnexpectedCharacterError(
                                    ch,
                                    self._pda.stack,
                                    self._pda.state,
                                    "Expected '[' or '\"' for 'StreamableValues'",
                                )
                        elif ch in ARRAY_OPEN:
                            self._pda.set_state(S.EXPECT_VALUE)
                            self._pda.push(M.ARRAY)
                        else:
                            raise UnexpectedCharacterError(
                                ch,
                                self._pda.stack,
                                self._pda.state,
                                "Expected '[' or white space.",
                            )

                    case S.PARSING_STRING:
                        if self._decoder.is_terminating_quote(ch):
                            if self._sink.current_sink_type == SinkType.AWAITABLE_VALUE:
                                MainType = self._sink.current_underlying_main_generic
                                if issubclass(MainType, Enum):
                                    try:
                                        value = MainType(self._decoder.buffer)
                                        await self._sink.emit(value)
                                    except ValueError as e:
                                        raise ParsePrimitiveError(
                                            f"Invalid enum value: "
                                            f"{self._decoder.buffer}"
                                        ) from e
                                else:
                                    await self._sink.emit(self._decoder.buffer)
                            self._decoder.reset()
                            await self._sink.close()
                            self._pda.set_state(S.EXPECT_COMMA_OR_EOC)
                        else:
                            maybe_char = self._decoder.push(ch)
                            if (
                                maybe_char is not None
                                and self._sink.current_sink_type
                                is SinkType.STREAMABLE_VALUES
                            ):
                                await self._sink.emit(maybe_char)

                    case _ if self._pda.state in PRIMITIVE_STATES:
                        if ch not in COMMA | OBJECT_CLOSE:
                            self._assert_primitive_character_allowed_in_state(ch)
                            self._decoder.push(ch)
                        else:
                            await self._parse_primitive()
                            await self._sink.close()
                            self._decoder.reset()
                            self._pda.set_state(S.EXPECT_KEY)
                            if ch in OBJECT_CLOSE:
                                await self._finalize()

                    case S.EXPECT_COMMA_OR_EOC:
                        if is_json_whitespace(ch):
                            pass
                        elif ch in COMMA:
                            self._pda.set_state(S.EXPECT_KEY)
                        elif ch in OBJECT_CLOSE:
                            await self._finalize()
                        else:
                            raise UnexpectedCharacterError(
                                ch,
                                self._pda.stack,
                                self._pda.state,
                                "Expected ',', '}' or white space.",
                            )

                    case _:
                        raise UnexpectedStateError(
                            self._pda.stack,
                            self._pda.state,
                            message="State not allowed in root context.",
                        )

            # CONTEXT: Array
            case M.ARRAY:
                if ch in ARRAY_OPEN:
                    raise UnexpectedCharacterError(
                        ch,
                        self._pda.stack,
                        self._pda.state,
                        "No support for 2-dimensional arrays.",
                    )

                match self._pda.state:
                    case S.EXPECT_VALUE:
                        if is_json_whitespace(ch):
                            pass
                        elif await self._handle_common__expect_value(ch):
                            pass
                        elif ch in ARRAY_CLOSE:
                            await self._close_context(S.EXPECT_COMMA_OR_EOC)
                        else:
                            raise UnexpectedCharacterError(
                                ch,
                                self._pda.stack,
                                self._pda.state,
                                "Expected value, ']' or white space",
                            )

                    case S.PARSING_STRING:
                        if self._sink.current_sink_type is SinkType.AWAITABLE_VALUE:
                            raise UnexpectedCharacterError(
                                ch,
                                self._pda.stack,
                                self._pda.state,
                                (
                                    "Cannot parse string inside of an array with "
                                    "AwaitableValue sink type."
                                ),
                            )
                        if self._decoder.is_terminating_quote(ch):
                            MainType = self._sink.current_underlying_main_generic
                            if issubclass(MainType, Enum):
                                try:
                                    value = MainType(self._decoder.buffer)
                                    await self._sink.emit(value)
                                except ValueError as e:
                                    raise ParsePrimitiveError(
                                        f"Invalid enum value: {self._decoder.buffer}"
                                    ) from e
                            else:
                                await self._sink.emit(self._decoder.buffer)
                            self._decoder.reset()
                            self._pda.set_state(S.EXPECT_COMMA_OR_EOC)
                        else:
                            self._decoder.push(ch)

                    case _ if self._pda.state in PRIMITIVE_STATES:
                        if ch not in COMMA | ARRAY_CLOSE:
                            self._assert_primitive_character_allowed_in_state(ch)
                            self._decoder.push(ch)
                        else:
                            await self._parse_primitive()
                            self._decoder.reset()
                            if ch in COMMA:
                                self._pda.set_state(S.EXPECT_VALUE)
                            elif ch in ARRAY_CLOSE:
                                await self._close_context(S.EXPECT_COMMA_OR_EOC)

                    case S.EXPECT_COMMA_OR_EOC:
                        if is_json_whitespace(ch):
                            pass
                        elif ch in COMMA:
                            self._pda.set_state(S.EXPECT_VALUE)
                        elif ch in ARRAY_CLOSE:
                            await self._close_context(S.EXPECT_COMMA_OR_EOC)
                        else:
                            raise UnexpectedCharacterError(
                                ch,
                                self._pda.stack,
                                self._pda.state,
                                "Expected ',', ']' or white space.",
                            )

                    case _:
                        raise UnexpectedStateError(
                            self._pda.stack,
                            self._pda.state,
                            message="State not allowed in array context.",
                        )

            # CONTEXT: Object
            case M.OBJECT:
                if self._pda.state is not S.PARSING_OBJECT:
                    raise UnexpectedCharacterError(
                        ch,
                        self._pda.stack,
                        self._pda.state,
                        "State in object context must be 'parsing_object'",
                    )
                if ch in OBJECT_CLOSE:
                    self._pda.pop()
                    if self._pda.top is M.ROOT:
                        await self._sink.close()
                    self._pda.set_state(S.EXPECT_COMMA_OR_EOC)
                    return
                else:
                    await self._sink.forward_char(ch)
                    return

    async def _parse_primitive(self) -> None:
        if self._pda.state is S.PARSING_NULL:
            if not self._decoder.buffer == JSON_NULL:
                raise ParsePrimitiveError(
                    f"Expected 'null', got '{self._decoder.buffer}'"
                )
            await self._sink.emit(None)
        elif self._pda.state is S.PARSING_BOOLEAN:
            await self._sink.emit(self._decoder.buffer == JSON_TRUE)
        else:
            try:
                buffer = self._decoder.buffer
                generic = self._sink.current_underlying_main_generic
                value = float(buffer) if issubclass(generic, float) else int(buffer)
            except ValueError as e:
                raise ParsePrimitiveError(f"Buffer: {buffer}; Error: {e}") from e
            await self._sink.emit(value)

    async def _handle_common__expect_value(self, ch: str) -> S | None:
        generic_set = self._sink.current_underlying_generics
        generic = self._sink.current_underlying_main_generic
        if ch in QUOTE:
            if not (str in generic_set or issubclass(generic, Enum)):
                raise UnexpectedCharacterError(
                    ch,
                    self._pda.stack,
                    self._pda.state,
                    (
                        "Trying to parse 'string' but underlying generic is "
                        f"'{generic_set}'."
                    ),
                )
            self._pda.set_state(S.PARSING_STRING)
            self._decoder.reset()
            return S.PARSING_STRING
        if ch in NUMBER_OPEN:
            if not any(t in generic_set for t in (int, float)):
                raise UnexpectedCharacterError(
                    ch,
                    self._pda.stack,
                    self._pda.state,
                    (
                        "Trying to parse 'number' but underlying generic is "
                        f"'{generic_set}'."
                    ),
                )
            self._decoder.push(ch)
            if generic is int:
                self._pda.set_state(S.PARSING_INTEGER)
                return S.PARSING_INTEGER
            else:
                self._pda.set_state(S.PARSING_FLOAT)
                return S.PARSING_FLOAT
        if ch in BOOLEAN_OPEN:
            if bool not in generic_set:
                raise UnexpectedCharacterError(
                    ch,
                    self._pda.stack,
                    self._pda.state,
                    (
                        "Trying to parse 'boolean' but underlying generic is "
                        f"'{generic.__name__}'."
                    ),
                )
            self._pda.set_state(S.PARSING_BOOLEAN)
            self._decoder.push(ch)
            return S.PARSING_BOOLEAN
        if ch in NULL_OPEN:
            if NoneType not in generic_set:
                raise UnexpectedCharacterError(
                    ch,
                    self._pda.stack,
                    self._pda.state,
                    (
                        "Trying to parse 'null' but underlying generic is "
                        f"'{generic.__name__}'."
                    ),
                )
            self._pda.set_state(S.PARSING_NULL)
            self._decoder.push(ch)
            return S.PARSING_NULL
        if ch in OBJECT_OPEN:
            if not issubclass(generic, JMux):
                raise UnexpectedCharacterError(
                    ch,
                    self._pda.stack,
                    self._pda.state,
                    (
                        f"Trying to parse 'object' but underlying generic is "
                        f"'{generic.__name__}'."
                    ),
                )
            await self._sink.create_and_emit_nested()
            await self._sink.forward_char(ch)
            self._pda.set_state(S.PARSING_OBJECT)
            self._pda.push(M.OBJECT)
            return S.PARSING_OBJECT

    async def _close_context(self, new_state: S) -> None:
        await self._sink.close()
        self._pda.pop()
        self._pda.set_state(new_state)

    async def _finalize(self) -> None:
        type_hints = get_type_hints(self.__class__)
        for attr_name, _ in type_hints.items():
            self._sink.set_current(attr_name)
            try:
                await self._sink.ensure_closed()
            except NothingEmittedError as e:
                raise NotAllObjectPropertiesSetError(
                    f"Unable to finalize. Property '{attr_name}' was not set before "
                    "closing the JMux instance."
                ) from e

        self._pda.pop()
        self._pda.set_state(S.END)

    def _assert_primitive_character_allowed_in_state(self, ch: str) -> None:
        if self._pda.state is S.PARSING_INTEGER:
            if ch not in INTERGER_ALLOWED:
                raise UnexpectedCharacterError(
                    ch,
                    self._pda.stack,
                    self._pda.state,
                    "Trying to parse 'integer' but received unexpected character.",
                )
        elif self._pda.state is S.PARSING_FLOAT:
            if ch not in FLOAT_ALLOWED:
                raise UnexpectedCharacterError(
                    ch,
                    self._pda.stack,
                    self._pda.state,
                    "Trying to parse 'float' but received unexpected character.",
                )
        elif self._pda.state is S.PARSING_BOOLEAN:
            if ch not in BOOLEAN_ALLOWED:
                raise UnexpectedCharacterError(
                    ch,
                    self._pda.stack,
                    self._pda.state,
                    "Trying to parse 'boolean' but received unexpected character.",
                )
            if not any(
                bool_str.startswith(f"{self._decoder.buffer}{ch}")
                for bool_str in (JSON_TRUE, JSON_FALSE)
            ):
                raise UnexpectedCharacterError(
                    ch,
                    self._pda.stack,
                    self._pda.state,
                    (
                        "Unexpected character added to buffer for 'boolean': "
                        f"'{self._decoder.buffer}{ch}'."
                    ),
                )
        elif self._pda.state is S.PARSING_NULL:
            if ch not in NULL_ALLOWED:
                raise UnexpectedCharacterError(
                    ch,
                    self._pda.stack,
                    self._pda.state,
                    "Trying to parse 'null' but received unexpected character.",
                )
            if not JSON_NULL.startswith(f"{self._decoder.buffer}{ch}"):
                raise UnexpectedCharacterError(
                    ch,
                    self._pda.stack,
                    self._pda.state,
                    (
                        f"Unexpected character added to buffer for 'null': "
                        f"'{self._decoder.buffer}{ch}'."
                    ),
                )
        else:
            raise UnexpectedCharacterError(
                ch,
                self._pda.stack,
                self._pda.state,
                "An unexpected error happened.",
            )



from types import NoneType, UnionType
from typing import Set, Tuple, Type, Union, get_args, get_origin


def is_json_whitespace(ch: str) -> bool:
    return ch in {" ", "\t", "\n", "\r"}


def extract_types_from_generic_alias(UnknownType: Type) -> Tuple[Set[Type], Set[Type]]:
    Origin: Type | None = get_origin(UnknownType)
    if Origin is None:
        return {UnknownType}, set()
    if Origin is UnionType or Origin is Union:
        return deconstruct_type(UnknownType), set()

    type_args = get_args(UnknownType)
    if len(type_args) != 1:
        raise TypeError(
            f"Only single type generics can be deconstruct with this function, "
            f"got {type_args}."
        )

    Generic: Type = type_args[0]
    type_set = deconstruct_type(Generic)
    if len(type_set) == 1:
        return {Origin}, type_set
    if len(type_set) != 2:
        raise TypeError(
            f"Union type must have exactly two types in its union, "
            f"got {get_args(Generic)}."
        )
    if NoneType not in get_args(Generic):
        raise TypeError(
            "Union type must include NoneType if it is used as a generic argument."
        )
    return {Origin}, type_set


def deconstruct_type(UnknownType: Type) -> Set[Type]:
    Origin: Type | None = get_origin(UnknownType)
    if UnknownType is None:
        return {NoneType}
    if Origin is None:
        return {UnknownType}
    if not (Origin is UnionType or Origin is Union):
        return {Origin}
    type_args = get_args(UnknownType)
    return set(type_args)


def get_main_type(type_set: Set[Type]) -> Type:
    type_set_copy: Set[Type] = type_set.copy()
    if NoneType in type_set_copy and len(type_set_copy) == 2:
        type_set_copy.remove(NoneType)
    if len(type_set_copy) != 1:
        raise TypeError(
            f"Expected exactly one type, got {type_set_copy}. If you want to allow "
            "NoneType, use Union[int, NoneType]."
        )
    return type_set_copy.pop()



from typing import List

import pytest

from jmux.decoder import StringEscapeDecoder


# fmt: off
@pytest.mark.parametrize(
    "stream,expected_string",
    [
        ("foo bar", "foo bar"),
        ("foo\"bar", 'foo"bar'),
        ("foo bar\"", 'foo bar"'),
        ("foo\\\\bar", "foo\\bar"),
        ("foo\\bbar", "foo\bbar"),
        ("foo\\tbar", "foo\tbar"),
        ("foo\\rbar", "foo\rbar"),
        ("foo\\fbar", "foo\fbar"),
        ("foo\\/bar", "foo/bar"),
        ("foo\\/bar", "foo/bar"),
    ],
)
# fmt: on
@pytest.mark.anyio
async def test_string_decoder__parameterized(stream: str, expected_string: List[str]):
    decoder = StringEscapeDecoder()

    for ch in stream:
        decoder.push(ch)

    assert decoder.buffer == expected_string



from typing import List


class PushDownAutomata[Context, State]:
    def __init__(self, start_state: State) -> None:
        self._stack: List[Context] = []
        self._state: State = start_state

    @property
    def state(self) -> State:
        return self._state

    @property
    def stack(self) -> List[Context]:
        return self._stack

    @property
    def top(self) -> Context | None:
        if not self._stack:
            return None
        return self._stack[-1]

    def set_state(self, new_state: State) -> None:
        self._state = new_state

    def push(self, mode: Context) -> None:
        self._stack.append(mode)

    def pop(self) -> Context:
        if not self._stack:
            raise IndexError("PDA stack is empty.")
        return self._stack.pop()



from types import NoneType
from typing import Set, Type

import pytest

from jmux.awaitable import (
    AwaitableValue,
    IAsyncSink,
    StreamableValues,
)
from jmux.demux import JMux


class NestedObject(JMux):
    key: AwaitableValue[str]


@pytest.mark.parametrize(
    "TargetType,expected_set",
    [
        (AwaitableValue[int], {int}),
        (AwaitableValue[float], {float}),
        (AwaitableValue[str], {str}),
        (AwaitableValue[bool], {bool}),
        (AwaitableValue[NestedObject], {NestedObject}),
        (AwaitableValue[int | None], {int, NoneType}),
        (AwaitableValue[float | None], {float, NoneType}),
        (AwaitableValue[str | None], {str, NoneType}),
        (AwaitableValue[bool | None], {bool, NoneType}),
        (AwaitableValue[NestedObject | None], {NestedObject, NoneType}),
        (StreamableValues[int], {int}),
        (StreamableValues[float], {float}),
        (StreamableValues[str], {str}),
        (StreamableValues[bool], {bool}),
        (StreamableValues[NestedObject], {NestedObject}),
    ],
)
def test_underlying_generic_mixin__get_underlying_generic__expected_set(
    TargetType: Type[IAsyncSink], expected_set: Set[Type]
):
    target = TargetType()
    underlying_types = target.get_underlying_generics()

    assert underlying_types == expected_set


@pytest.mark.parametrize(
    "TargetType,MaybeExpectedError",
    [
        (AwaitableValue[int], None),
        (AwaitableValue[str], None),
        (AwaitableValue[float], None),
        (AwaitableValue[bool], None),
        (AwaitableValue[NoneType], None),
        (AwaitableValue[NestedObject | None], None),
        (AwaitableValue[int | NoneType], None),
        (AwaitableValue[str | NoneType], None),
        (AwaitableValue[float | NoneType], None),
        (AwaitableValue[bool | NoneType], None),
        (AwaitableValue[bool | str | NoneType], TypeError),
        (AwaitableValue[bool | str], TypeError),
        (AwaitableValue[NestedObject | str], TypeError),
        (StreamableValues[int | None], TypeError),
        (StreamableValues[float | None], TypeError),
        (StreamableValues[str | None], TypeError),
        (StreamableValues[bool | None], TypeError),
        (StreamableValues[NestedObject | None], TypeError),
    ],
)
def test_underlying_generic_mixin__get_underlying_generic__check_instantiation(
    TargetType: Type[IAsyncSink], MaybeExpectedError: Type[Exception] | None
):
    target = TargetType()
    if MaybeExpectedError:
        with pytest.raises(MaybeExpectedError):
            _ = target.get_underlying_generics()
    else:
        _ = target.get_underlying_generics()



from enum import Enum
from types import NoneType
from typing import List, Type

import pytest

from jmux.awaitable import AwaitableValue, StreamableValues
from jmux.demux import JMux
from jmux.error import (
    EmptyKeyError,
    MissingAttributeError,
    ObjectAlreadyClosedError,
    ParsePrimitiveError,
    UnexpectedCharacterError,
)
from jmux.types import Mode, State


# fmt: off
@pytest.mark.parametrize(
    "stream,expected_stack,expected_state",
    [
        ("", [], State.START),
        ("{", [Mode.ROOT], State.EXPECT_KEY),
        ("{ ", [Mode.ROOT], State.EXPECT_KEY),
        ('{"', [Mode.ROOT], State.PARSING_KEY),
        ('{"key_', [Mode.ROOT], State.PARSING_KEY),
        ('{"key_str', [Mode.ROOT], State.PARSING_KEY),
        ('{"key_str"', [Mode.ROOT], State.EXPECT_COLON),
        ('{"key_str":', [Mode.ROOT], State.EXPECT_VALUE),
        ('{"key_str": ', [Mode.ROOT], State.EXPECT_VALUE),
        ('{"key_str": \t\n', [Mode.ROOT], State.EXPECT_VALUE),
        ('{"key_str": "', [Mode.ROOT], State.PARSING_STRING),
        ('{"key_str": "val', [Mode.ROOT], State.PARSING_STRING),
        ('{"key_str": "val"', [Mode.ROOT], State.EXPECT_COMMA_OR_EOC),
        ('{"key_str": "val" \t\n', [Mode.ROOT], State.EXPECT_COMMA_OR_EOC),
        ('{"key_str": "val",', [Mode.ROOT], State.EXPECT_KEY),
        ('{"key_str": "val","key_int', [Mode.ROOT], State.PARSING_KEY),
        ('{"key_str": "val","key_int"', [Mode.ROOT], State.EXPECT_COLON),
        ('{"key_str": "val","key_int":', [Mode.ROOT], State.EXPECT_VALUE),
        ('{"key_str": "val","key_int": \t\n', [Mode.ROOT], State.EXPECT_VALUE),
        ('{"key_str": "val","key_int":4', [Mode.ROOT], State.PARSING_INTEGER),
        ('{"key_str": "val","key_int":42', [Mode.ROOT], State.PARSING_INTEGER),
        ('{"key_str": "val","key_int":42,', [Mode.ROOT], State.EXPECT_KEY),
        ('{"key_str": "val","key_int":42,"', [Mode.ROOT], State.PARSING_KEY),
        ('{"key_str": "val","key_int":42,"key_float"', [Mode.ROOT], State.EXPECT_COLON),
        ('{"key_str": "val","key_int":42,"key_float":', [Mode.ROOT], State.EXPECT_VALUE),
        ('{"key_str": "val","key_int":42,"key_float":', [Mode.ROOT], State.EXPECT_VALUE),
        ('{"key_str": "val","key_int":42,"key_float":3.14', [Mode.ROOT], State.PARSING_FLOAT),
        ('{"key_str": "val","key_int":42,"key_float":3.14,', [Mode.ROOT], State.EXPECT_KEY),
        ('{"key_str": "val","key_int":42,"key_float":3.14,"key_bool":', [Mode.ROOT], State.EXPECT_VALUE),
        ('{"key_str": "val","key_int":42,"key_float":3.14,"key_bool":t', [Mode.ROOT], State.PARSING_BOOLEAN),
        ('{"key_str": "val","key_int":42,"key_float":3.14,"key_bool":true', [Mode.ROOT], State.PARSING_BOOLEAN),
        ('{"key_str": "val","key_int":42,"key_float":3.14,"key_bool":true,"key_none":n', [Mode.ROOT], State.PARSING_NULL),
        ('{"key_str": "val","key_int":42,"key_float":3.14,"key_bool":true,"key_none":null,', [Mode.ROOT], State.EXPECT_KEY),
        ('{"key_str": "val","key_int":42,"key_float":3.14,"key_bool":true,"key_none":null,', [Mode.ROOT], State.EXPECT_KEY),
        ('{"key_str": "val","key_int":42,"key_float":3.14,"key_bool":true,"key_none":null,"key_stream', [Mode.ROOT], State.PARSING_KEY),
        ('{"key_str": "val","key_int":42,"key_float":3.14,"key_bool":true,"key_none":null,"key_stream":"stream', [Mode.ROOT], State.PARSING_STRING),
        ('{"key_str": "val","key_int":42,"key_float":3.14,"key_bool":true,"key_none":null,"key_stream":"stream","key_enum', [Mode.ROOT], State.PARSING_KEY),
        ('{"key_str": "val","key_int":42,"key_float":3.14,"key_bool":true,"key_none":null,"key_stream":"stream","key_enum":"val', [Mode.ROOT], State.PARSING_STRING),
        ('{"key_str": "val","key_int":42,"key_float":3.14,"key_bool":true,"key_none":null,"key_stream":"stream","key_enum":"value1"', [Mode.ROOT], State.EXPECT_COMMA_OR_EOC),
        ('{"key_str": "val","key_int":42,"key_float":3.14,"key_bool":true,"key_none":null,"key_stream":"stream","key_enum":"value1","key_nested":', [Mode.ROOT], State.EXPECT_VALUE),
        ('{"key_str": "val","key_int":42,"key_float":3.14,"key_bool":true,"key_none":null,"key_stream":"stream","key_enum":"value1","key_nested":{', [Mode.ROOT, Mode.OBJECT], State.PARSING_OBJECT),
        ('{"key_str": "val","key_int":42,"key_float":3.14,"key_bool":true,"key_none":null,"key_stream":"stream","key_enum":"value1","key_nested":{"', [Mode.ROOT, Mode.OBJECT], State.PARSING_OBJECT),
        ('{"key_str": "val","key_int":42,"key_float":3.14,"key_bool":true,"key_none":null,"key_stream":"stream","key_enum":"value1","key_nested":{"key_str"', [Mode.ROOT, Mode.OBJECT], State.PARSING_OBJECT),
        ('{"key_str": "val","key_int":42,"key_float":3.14,"key_bool":true,"key_none":null,"key_stream":"stream","key_enum":"value1","key_nested":{"key_str":"nested"', [Mode.ROOT, Mode.OBJECT], State.PARSING_OBJECT),
        ('{"key_str": "val","key_int":42,"key_float":3.14,"key_bool":true,"key_none":null,"key_stream":"stream","key_enum":"value1","key_nested":{"key_str":"nested"}', [Mode.ROOT], State.EXPECT_COMMA_OR_EOC),
        ('{"key_str": "val","key_int":42,"key_float":3.14,"key_bool":true,"key_none":null,"key_stream":"stream","key_enum":"value1","key_nested":{"key_str":"nested"},', [Mode.ROOT], State.EXPECT_KEY),
        ('{"key_str": "val","key_int":42,"key_float":3.14,"key_bool":true,"key_none":null,"key_stream":"stream","key_enum":"value1","key_nested":{"key_str":"nested"},"arr_str":', [Mode.ROOT], State.EXPECT_VALUE),
        ('{"key_str": "val","key_int":42,"key_float":3.14,"key_bool":true,"key_none":null,"key_stream":"stream","key_enum":"value1","key_nested":{"key_str":"nested"},"arr_str":[', [Mode.ROOT, Mode.ARRAY], State.EXPECT_VALUE),
        ('{"key_str": "val","key_int":42,"key_float":3.14,"key_bool":true,"key_none":null,"key_stream":"stream","key_enum":"value1","key_nested":{"key_str":"nested"},"arr_str":["', [Mode.ROOT, Mode.ARRAY], State.PARSING_STRING),
        ('{"key_str": "val","key_int":42,"key_float":3.14,"key_bool":true,"key_none":null,"key_stream":"stream","key_enum":"value1","key_nested":{"key_str":"nested"},"arr_str":["val1"', [Mode.ROOT, Mode.ARRAY], State.EXPECT_COMMA_OR_EOC),
        ('{"key_str": "val","key_int":42,"key_float":3.14,"key_bool":true,"key_none":null,"key_stream":"stream","key_enum":"value1","key_nested":{"key_str":"nested"},"arr_str":["val1" \t\n', [Mode.ROOT, Mode.ARRAY], State.EXPECT_COMMA_OR_EOC),
        ('{"key_str": "val","key_int":42,"key_float":3.14,"key_bool":true,"key_none":null,"key_stream":"stream","key_enum":"value1","key_nested":{"key_str":"nested"},"arr_str":["val1",', [Mode.ROOT, Mode.ARRAY], State.EXPECT_VALUE),
        ('{"key_str": "val","key_int":42,"key_float":3.14,"key_bool":true,"key_none":null,"key_stream":"stream","key_enum":"value1","key_nested":{"key_str":"nested"},"arr_str":["val1", \t\n', [Mode.ROOT, Mode.ARRAY], State.EXPECT_VALUE),
        ('{"key_str": "val","key_int":42,"key_float":3.14,"key_bool":true,"key_none":null,"key_stream":"stream","key_enum":"value1","key_nested":{"key_str":"nested"},"arr_str":["val1","val2",', [Mode.ROOT, Mode.ARRAY], State.EXPECT_VALUE),
        ('{"key_str": "val","key_int":42,"key_float":3.14,"key_bool":true,"key_none":null,"key_stream":"stream","key_enum":"value1","key_nested":{"key_str":"nested"},"arr_str":["val1","val2","val3', [Mode.ROOT, Mode.ARRAY], State.PARSING_STRING),
        ('{"key_str": "val","key_int":42,"key_float":3.14,"key_bool":true,"key_none":null,"key_stream":"stream","key_enum":"value1","key_nested":{"key_str":"nested"},"arr_str":["val1","val2","val3"', [Mode.ROOT, Mode.ARRAY], State.EXPECT_COMMA_OR_EOC),
        ('{"key_str": "val","key_int":42,"key_float":3.14,"key_bool":true,"key_none":null,"key_stream":"stream","key_enum":"value1","key_nested":{"key_str":"nested"},"arr_str":["val1","val2","val3"]', [Mode.ROOT], State.EXPECT_COMMA_OR_EOC),
        ('{"key_str": "val","key_int":42,"key_float":3.14,"key_bool":true,"key_none":null,"key_stream":"stream","key_enum":"value1","key_nested":{"key_str":"nested"},"arr_str":["val1","val2","val3"],', [Mode.ROOT], State.EXPECT_KEY),
        ('{"key_str": "val","key_int":42,"key_float":3.14,"key_bool":true,"key_none":null,"key_stream":"stream","key_enum":"value1","key_nested":{"key_str":"nested"},"arr_str":["val1","val2","val3"],"arr_int":[', [Mode.ROOT, Mode.ARRAY], State.EXPECT_VALUE),
        ('{"key_str": "val","key_int":42,"key_float":3.14,"key_bool":true,"key_none":null,"key_stream":"stream","key_enum":"value1","key_nested":{"key_str":"nested"},"arr_str":["val1","val2","val3"],"arr_int":[42', [Mode.ROOT, Mode.ARRAY], State.PARSING_INTEGER),
        ('{"key_str": "val","key_int":42,"key_float":3.14,"key_bool":true,"key_none":null,"key_stream":"stream","key_enum":"value1","key_nested":{"key_str":"nested"},"arr_str":["val1","val2","val3"],"arr_int":[42,', [Mode.ROOT, Mode.ARRAY], State.EXPECT_VALUE),
        ('{"key_str": "val","key_int":42,"key_float":3.14,"key_bool":true,"key_none":null,"key_stream":"stream","key_enum":"value1","key_nested":{"key_str":"nested"},"arr_str":["val1","val2","val3"],"arr_int":[42,43],"arr_float":[3', [Mode.ROOT, Mode.ARRAY], State.PARSING_FLOAT),
        ('{"key_str": "val","key_int":42,"key_float":3.14,"key_bool":true,"key_none":null,"key_stream":"stream","key_enum":"value1","key_nested":{"key_str":"nested"},"arr_str":["val1","val2","val3"],"arr_int":[42,43],"arr_float":[3.14', [Mode.ROOT, Mode.ARRAY], State.PARSING_FLOAT),
        ('{"key_str": "val","key_int":42,"key_float":3.14,"key_bool":true,"key_none":null,"key_stream":"stream","key_enum":"value1","key_nested":{"key_str":"nested"},"arr_str":["val1","val2","val3"],"arr_int":[42,43],"arr_float":[3.14,', [Mode.ROOT, Mode.ARRAY], State.EXPECT_VALUE),
        ('{"key_str": "val","key_int":42,"key_float":3.14,"key_bool":true,"key_none":null,"key_stream":"stream","key_enum":"value1","key_nested":{"key_str":"nested"},"arr_str":["val1","val2","val3"],"arr_int":[42,43],"arr_float":[3.14,31.4]', [Mode.ROOT], State.EXPECT_COMMA_OR_EOC),
        ('{"key_str": "val","key_int":42,"key_float":3.14,"key_bool":true,"key_none":null,"key_stream":"stream","key_enum":"value1","key_nested":{"key_str":"nested"},"arr_str":["val1","val2","val3"],"arr_int":[42,43],"arr_float":[3.14,31.4],"arr_bool":[true', [Mode.ROOT, Mode.ARRAY], State.PARSING_BOOLEAN),
        ('{"key_str": "val","key_int":42,"key_float":3.14,"key_bool":true,"key_none":null,"key_stream":"stream","key_enum":"value1","key_nested":{"key_str":"nested"},"arr_str":["val1","val2","val3"],"arr_int":[42,43],"arr_float":[3.14,31.4],"arr_bool":[true,false', [Mode.ROOT, Mode.ARRAY], State.PARSING_BOOLEAN),
        ('{"key_str": "val","key_int":42,"key_float":3.14,"key_bool":true,"key_none":null,"key_stream":"stream","key_enum":"value1","key_nested":{"key_str":"nested"},"arr_str":["val1","val2","val3"],"arr_int":[42,43],"arr_float":[3.14,31.4],"arr_bool":[true,false,true]', [Mode.ROOT], State.EXPECT_COMMA_OR_EOC),
        ('{"key_str": "val","key_int":42,"key_float":3.14,"key_bool":true,"key_none":null,"key_stream":"stream","key_enum":"value1","key_nested":{"key_str":"nested"},"arr_str":["val1","val2","val3"],"arr_int":[42,43],"arr_float":[3.14,31.4],"arr_bool":[true,false,true],"arr_none":[null,nul', [Mode.ROOT, Mode.ARRAY], State.PARSING_NULL),
        ('{"key_str": "val","key_int":42,"key_float":3.14,"key_bool":true,"key_none":null,"key_stream":"stream","key_enum":"value1","key_nested":{"key_str":"nested"},"arr_str":["val1","val2","val3"],"arr_int":[42,43],"arr_float":[3.14,31.4],"arr_bool":[true,false,true],"arr_none":[null,null]', [Mode.ROOT], State.EXPECT_COMMA_OR_EOC),
        ('{"key_str": "val","key_int":42,"key_float":3.14,"key_bool":true,"key_none":null,"key_stream":"stream","key_enum":"value1","key_nested":{"key_str":"nested"},"arr_str":["val1","val2","val3"],"arr_int":[42,43],"arr_float":[3.14,31.4],"arr_bool":[true,false,true],"arr_none":[null,null],"arr_enum":[', [Mode.ROOT, Mode.ARRAY], State.EXPECT_VALUE),
        ('{"key_str": "val","key_int":42,"key_float":3.14,"key_bool":true,"key_none":null,"key_stream":"stream","key_enum":"value1","key_nested":{"key_str":"nested"},"arr_str":["val1","val2","val3"],"arr_int":[42,43],"arr_float":[3.14,31.4],"arr_bool":[true,false,true],"arr_none":[null,null],"arr_enum":["val', [Mode.ROOT, Mode.ARRAY], State.PARSING_STRING),
        ('{"key_str": "val","key_int":42,"key_float":3.14,"key_bool":true,"key_none":null,"key_stream":"stream","key_enum":"value1","key_nested":{"key_str":"nested"},"arr_str":["val1","val2","val3"],"arr_int":[42,43],"arr_float":[3.14,31.4],"arr_bool":[true,false,true],"arr_none":[null,null],"arr_enum":["value1"', [Mode.ROOT, Mode.ARRAY], State.EXPECT_COMMA_OR_EOC),
        ('{"key_str": "val","key_int":42,"key_float":3.14,"key_bool":true,"key_none":null,"key_stream":"stream","key_enum":"value1","key_nested":{"key_str":"nested"},"arr_str":["val1","val2","val3"],"arr_int":[42,43],"arr_float":[3.14,31.4],"arr_bool":[true,false,true],"arr_none":[null,null],"arr_enum":["value1","value2"', [Mode.ROOT, Mode.ARRAY], State.EXPECT_COMMA_OR_EOC),
        ('{"key_str": "val","key_int":42,"key_float":3.14,"key_bool":true,"key_none":null,"key_stream":"stream","key_enum":"value1","key_nested":{"key_str":"nested"},"arr_str":["val1","val2","val3"],"arr_int":[42,43],"arr_float":[3.14,31.4],"arr_bool":[true,false,true],"arr_none":[null,null],"arr_enum":["value1","value2"]', [Mode.ROOT], State.EXPECT_COMMA_OR_EOC),
        ('{"key_str": "val","key_int":42,"key_float":3.14,"key_bool":true,"key_none":null,"key_stream":"stream","key_enum":"value1","key_nested":{"key_str":"nested"},"arr_str":["val1","val2","val3"],"arr_int":[42,43],"arr_float":[3.14,31.4],"arr_bool":[true,false,true],"arr_none":[null,null],"arr_enum":["value1","value2"],"arr_nested":[{', [Mode.ROOT, Mode.ARRAY, Mode.OBJECT], State.PARSING_OBJECT),
        ('{"key_str": "val","key_int":42,"key_float":3.14,"key_bool":true,"key_none":null,"key_stream":"stream","key_enum":"value1","key_nested":{"key_str":"nested"},"arr_str":["val1","val2","val3"],"arr_int":[42,43],"arr_float":[3.14,31.4],"arr_bool":[true,false,true],"arr_none":[null,null],"arr_enum":["value1","value2"],"arr_nested":[{"key_s', [Mode.ROOT, Mode.ARRAY, Mode.OBJECT], State.PARSING_OBJECT),
        ('{"key_str": "val","key_int":42,"key_float":3.14,"key_bool":true,"key_none":null,"key_stream":"stream","key_enum":"value1","key_nested":{"key_str":"nested"},"arr_str":["val1","val2","val3"],"arr_int":[42,43],"arr_float":[3.14,31.4],"arr_bool":[true,false,true],"arr_none":[null,null],"arr_enum":["value1","value2"],"arr_nested":[{"key_str":"nested1"}', [Mode.ROOT, Mode.ARRAY], State.EXPECT_COMMA_OR_EOC),
        ('{"key_str": "val","key_int":42,"key_float":3.14,"key_bool":true,"key_none":null,"key_stream":"stream","key_enum":"value1","key_nested":{"key_str":"nested"},"arr_str":["val1","val2","val3"],"arr_int":[42,43],"arr_float":[3.14,31.4],"arr_bool":[true,false,true],"arr_none":[null,null],"arr_enum":["value1","value2"],"arr_nested":[{"key_str":"nested1"},', [Mode.ROOT, Mode.ARRAY], State.EXPECT_VALUE),
        ('{"key_str": "val","key_int":42,"key_float":3.14,"key_bool":true,"key_none":null,"key_stream":"stream","key_enum":"value1","key_nested":{"key_str":"nested"},"arr_str":["val1","val2","val3"],"arr_int":[42,43],"arr_float":[3.14,31.4],"arr_bool":[true,false,true],"arr_none":[null,null],"arr_enum":["value1","value2"],"arr_nested":[{"key_str":"nested1"},{"key_str":"nes', [Mode.ROOT, Mode.ARRAY, Mode.OBJECT], State.PARSING_OBJECT),
        ('{"key_str": "val","key_int":42,"key_float":3.14,"key_bool":true,"key_none":null,"key_stream":"stream","key_enum":"value1","key_nested":{"key_str":"nested"},"arr_str":["val1","val2","val3"],"arr_int":[42,43],"arr_float":[3.14,31.4],"arr_bool":[true,false,true],"arr_none":[null,null],"arr_enum":["value1","value2"],"arr_nested":[{"key_str":"nested1"},{"key_str":"nested2"}', [Mode.ROOT, Mode.ARRAY], State.EXPECT_COMMA_OR_EOC),
        ('{"key_str": "val","key_int":42,"key_float":3.14,"key_bool":true,"key_none":null,"key_stream":"stream","key_enum":"value1","key_nested":{"key_str":"nested"},"arr_str":["val1","val2","val3"],"arr_int":[42,43],"arr_float":[3.14,31.4],"arr_bool":[true,false,true],"arr_none":[null,null],"arr_enum":["value1","value2"],"arr_nested":[{"key_str":"nested1"},{"key_str":"nested2"}]', [Mode.ROOT], State.EXPECT_COMMA_OR_EOC),
        ('{"key_str": "val","key_int":42,"key_float":3.14,"key_bool":true,"key_none":null,"key_stream":"stream","key_enum":"value1","key_nested":{"key_str":"nested"},"arr_str":["val1","val2","val3"],"arr_int":[42,43],"arr_float":[3.14,31.4],"arr_bool":[true,false,true],"arr_none":[null,null],"arr_enum":["value1","value2"],"arr_nested":[{"key_str":"nested1"},{"key_str":"nested2"}]}', [], State.END),
    ],
)
# fmt: on
@pytest.mark.anyio
async def test_json_demux__parse_correct_stream__assert_state(
    stream: str, expected_stack: List[Mode], expected_state: State
):
    class SObject(JMux):
        class SNested(JMux):
            key_str: AwaitableValue[str]

        class SEnum(Enum):
            VALUE1 = "value1"
            VALUE2 = "value2"

        key_str: AwaitableValue[str]
        key_int: AwaitableValue[int]
        key_float: AwaitableValue[float]
        key_bool: AwaitableValue[bool]
        key_none: AwaitableValue[NoneType]
        key_stream: StreamableValues[str]
        key_enum: AwaitableValue[SEnum]
        key_nested: AwaitableValue[SNested]

        arr_str: StreamableValues[str]
        arr_int: StreamableValues[int]
        arr_float: StreamableValues[float]
        arr_bool: StreamableValues[bool]
        arr_none: StreamableValues[NoneType]
        arr_enum: StreamableValues[SEnum]
        arr_nested: StreamableValues[SNested]

    s_object = SObject()

    for ch in stream:
        await s_object.feed_char(ch)

    assert s_object._pda.state == expected_state
    assert s_object._pda._stack == expected_stack


# fmt: off
@pytest.mark.parametrize(
    "stream,MaybeExpectedError",
    [
        ("b", UnexpectedCharacterError),
        ("{", None),
        ("{p", UnexpectedCharacterError),
        ('{"', None),
        ('{""', EmptyKeyError),
        ('{"no_actual_key"', MissingAttributeError),
        ('{"key_str"', None),
        ('{"key_str": ""', None),
        ('{"key_str": "val","key_int":4p', UnexpectedCharacterError),
        ('{"key_str": "val","key_int":4t', UnexpectedCharacterError),
        ('{"key_str": "val","key_int":420', None),
        ('{"key_str": "val","key_int":-420', None),
        ('{"key_str": "val","key_int":-4.20', UnexpectedCharacterError),
        ('{"key_str": "val","key_int":42,"key_float":1e+', None),
        ('{"key_str": "val","key_int":42,"key_float":0', None),
        ('{"key_str": "val","key_int":42,"key_float":p', UnexpectedCharacterError),
        ('{"key_str": "val","key_int":42,"key_float":1e+,', ParsePrimitiveError),
        ('{"key_str": "val","key_int":42,"key_float":-3.14e10,', None),
        ('{"key_str": "val","key_int":42,"key_float":-2.5E3,', None),
        ('{"key_str": "val","key_int":42,"key_float":1E+10,', None),
        ('{"key_str": "val","key_int":42,"key_float":NaN', UnexpectedCharacterError),
        ('{"key_str": "val","key_int":42,"key_float":Infinity', UnexpectedCharacterError),
        ('{"key_str": "val","key_int":42,"key_float":-', None),
        ('{"key_str": "val","key_int":42,"key_float":+', UnexpectedCharacterError),
        ('{"key_str": "val","key_int":42,"key_float":-1', None),
        ('{"key_str": "val","key_int":42,"key_float":--1', None),
        ('{"key_str": "val","key_int":42,"key_float":--1,', ParsePrimitiveError),
        ('{"key_str": "val","key_int":42,"key_float":.', UnexpectedCharacterError),
        ('{"key_str": "val","key_int":42,"key_float":1.', None),
        ('{"key_str": "val","key_int":42,"key_float":3.14,"key_bool":t', None),
        ('{"key_str": "val","key_int":42,"key_float":3.14,"key_bool":T', UnexpectedCharacterError),
        ('{"key_str": "val","key_int":42,"key_float":3.14,"key_bool":trub', UnexpectedCharacterError),
        ('{"key_str": "val","key_int":42,"key_float":3.14,"key_bool":tf', UnexpectedCharacterError),
        ('{"key_str": "val","key_int":42,"key_float":3.14,"key_bool":trueee', UnexpectedCharacterError),
        ('{"key_str": "val","key_int":42,"key_float":3.14,"key_bool":true,', None),
        ('{"key_str": "val","key_int":42,"key_float":3.14,"key_bool":f', None),
        ('{"key_str": "val","key_int":42,"key_float":3.14,"key_bool":F', UnexpectedCharacterError),
        ('{"key_str": "val","key_int":42,"key_float":3.14,"key_bool":ft', UnexpectedCharacterError),
        ('{"key_str": "val","key_int":42,"key_float":3.14,"key_bool":falsb', UnexpectedCharacterError),
        ('{"key_str": "val","key_int":42,"key_float":3.14,"key_bool":false', None),
        ('{"key_str": "val","key_int":42,"key_float":3.14,"key_bool":false,', None),
        ('{"key_str": "val","key_int":42,"key_float":3.14,"key_bool":false,"key_none":', None),
        ('{"key_str": "val","key_int":42,"key_float":3.14,"key_bool":false,"key_none":n', None),
        ('{"key_str": "val","key_int":42,"key_float":3.14,"key_bool":false,"key_none":nope', UnexpectedCharacterError),
        ('{"key_str": "val","key_int":42,"key_float":3.14,"key_bool":false,"key_none":nulll', UnexpectedCharacterError),
        ('{"key_str": "val","key_int":42,"key_float":3.14,"key_bool":false,"key_none":null', None),
        ('{"key_str": "val","key_int":42,"key_float":3.14,"key_bool":false,"key_none":null,', None),
        ('{"key_str": "val","key_int":42,"key_float":3.14,"key_bool":false,"key_none":null,"key_enum', None),
        ('{"key_str": "val","key_int":42,"key_float":3.14,"key_bool":false,"key_none":null,"key_enum":,', UnexpectedCharacterError),
        ('{"key_str": "val","key_int":42,"key_float":3.14,"key_bool":false,"key_none":null,"key_enum":"val', None),
        ('{"key_str": "val","key_int":42,"key_float":3.14,"key_bool":false,"key_none":null,"key_enum":"foo', None),
        ('{"key_str": "val","key_int":42,"key_float":3.14,"key_bool":false,"key_none":null,"key_enum":"value1"', None),
        ('{"key_str": "val","key_int":42,"key_float":3.14,"key_bool":false,"key_none":null,"key_enum":"foobar"', ParsePrimitiveError),
        ('{"key_str": "val","key_int":42,"key_float":3.14,"key_bool":false,"key_none":null,"key_enum":"value1","key_nested":p', UnexpectedCharacterError),
        ('{"key_str": "val","key_int":42,"key_float":3.14,"key_bool":false,"key_none":null,"key_enum":"value1","key_nested":n', UnexpectedCharacterError),
        ('{"key_str": "val","key_int":42,"key_float":3.14,"key_bool":false,"key_none":null,"key_enum":"value1","key_nested":4', UnexpectedCharacterError),
        ('{"key_str": "val","key_int":42,"key_float":3.14,"key_bool":false,"key_none":null,"key_enum":"value1","key_nested":{', None),
        ('{"key_str": "val","key_int":42,"key_float":3.14,"key_bool":false,"key_none":null,"key_enum":"value1","key_nested":{p', UnexpectedCharacterError), # Means all recursive calls throw errors as expected
        ('{"key_str": "val","key_int":42,"key_float":3.14,"key_bool":false,"key_none":null,"key_enum":"value1","key_nested":{"key_str":"nested"},', None),
        ('{"key_str": "val","key_int":42,"key_float":3.14,"key_bool":false,"key_none":null,"key_enum":"value1","key_nested":{"key_str":"nested"},', None),
        ('{"key_str": "val","key_int":42,"key_float":3.14,"key_bool":false,"key_none":null,"key_enum":"value1","key_nested":{"key_str":"nested"},"arr_str":{', UnexpectedCharacterError),
        ('{"key_str": "val","key_int":42,"key_float":3.14,"key_bool":false,"key_none":null,"key_enum":"value1","key_nested":{"key_str":"nested"},"arr_str":p', UnexpectedCharacterError),
        ('{"key_str": "val","key_int":42,"key_float":3.14,"key_bool":false,"key_none":null,"key_enum":"value1","key_nested":{"key_str":"nested"},"arr_str":[[', UnexpectedCharacterError),
        ('{"key_str": "val","key_int":42,"key_float":3.14,"key_bool":false,"key_none":null,"key_enum":"value1","key_nested":{"key_str":"nested"},"arr_str":[]', None), # Allow empty arrays
        ('{"key_str": "val","key_int":42,"key_float":3.14,"key_bool":false,"key_none":null,"key_enum":"value1","key_nested":{"key_str":"nested"},"arr_str":[nu', UnexpectedCharacterError),
        ('{"key_str": "val","key_int":42,"key_float":3.14,"key_bool":false,"key_none":null,"key_enum":"value1","key_nested":{"key_str":"nested"},"arr_str":["', None),
        ('{"key_str": "val","key_int":42,"key_float":3.14,"key_bool":false,"key_none":null,"key_enum":"value1","key_nested":{"key_str":"nested"},"arr_str":["val1",}', UnexpectedCharacterError),
        ('{"key_str": "val","key_int":42,"key_float":3.14,"key_bool":false,"key_none":null,"key_enum":"value1","key_nested":{"key_str":"nested"},"arr_str":["val1"]', None),
        ('{"key_str": "val","key_int":42,"key_float":3.14,"key_bool":false,"key_none":null,"key_enum":"value1","key_nested":{"key_str":"nested"},"arr_str":["val1"],"arr_int":4,', UnexpectedCharacterError),
        ('{"key_str": "val","key_int":42,"key_float":3.14,"key_bool":false,"key_none":null,"key_enum":"value1","key_nested":{"key_str":"nested"},"arr_str":["val1"],"arr_int":[4.', UnexpectedCharacterError),
        ('{"key_str": "val","key_int":42,"key_float":3.14,"key_bool":false,"key_none":null,"key_enum":"value1","key_nested":{"key_str":"nested"},"arr_str":["val1"],"arr_int":[]', None),
        ('{"key_str": "val","key_int":42,"key_float":3.14,"key_bool":false,"key_none":null,"key_enum":"value1","key_nested":{"key_str":"nested"},"arr_str":["val1"],"arr_int":[42,', None),
        ('{"key_str": "val","key_int":42,"key_float":3.14,"key_bool":false,"key_none":null,"key_enum":"value1","key_nested":{"key_str":"nested"},"arr_str":["val1"],"arr_int":[42,[', UnexpectedCharacterError),
        ('{"key_str": "val","key_int":42,"key_float":3.14,"key_bool":false,"key_none":null,"key_enum":"value1","key_nested":{"key_str":"nested"},"arr_str":["val1"],"arr_int":[-42,', None),
        ('{"key_str": "val","key_int":42,"key_float":3.14,"key_bool":false,"key_none":null,"key_enum":"value1","key_nested":{"key_str":"nested"},"arr_str":["val1"],"arr_int":[42,+43]', UnexpectedCharacterError),
        ('{"key_str": "val","key_int":42,"key_float":3.14,"key_bool":false,"key_none":null,"key_enum":"value1","key_nested":{"key_str":"nested"},"arr_str":["val1"],"arr_int":[42,-43]', None),
        ('{"key_str": "val","key_int":42,"key_float":3.14,"key_bool":false,"key_none":null,"key_enum":"value1","key_nested":{"key_str":"nested"},"arr_str":["val1"],"arr_int":[42],"arr_float":3', UnexpectedCharacterError),
        ('{"key_str": "val","key_int":42,"key_float":3.14,"key_bool":false,"key_none":null,"key_enum":"value1","key_nested":{"key_str":"nested"},"arr_str":["val1"],"arr_int":[42],"arr_float":{', UnexpectedCharacterError),
        ('{"key_str": "val","key_int":42,"key_float":3.14,"key_bool":false,"key_none":null,"key_enum":"value1","key_nested":{"key_str":"nested"},"arr_str":["val1"],"arr_int":[42],"arr_float":"', UnexpectedCharacterError),
        ('{"key_str": "val","key_int":42,"key_float":3.14,"key_bool":false,"key_none":null,"key_enum":"value1","key_nested":{"key_str":"nested"},"arr_str":["val1"],"arr_int":[42],"arr_float":[3k', UnexpectedCharacterError),
        ('{"key_str": "val","key_int":42,"key_float":3.14,"key_bool":false,"key_none":null,"key_enum":"value1","key_nested":{"key_str":"nested"},"arr_str":["val1"],"arr_int":[42],"arr_float":[0', None),
        ('{"key_str": "val","key_int":42,"key_float":3.14,"key_bool":false,"key_none":null,"key_enum":"value1","key_nested":{"key_str":"nested"},"arr_str":["val1"],"arr_int":[42],"arr_float":[]', None),
        ('{"key_str": "val","key_int":42,"key_float":3.14,"key_bool":false,"key_none":null,"key_enum":"value1","key_nested":{"key_str":"nested"},"arr_str":["val1"],"arr_int":[42],"arr_float":[3.14,314]', None),
        ('{"key_str": "val","key_int":42,"key_float":3.14,"key_bool":false,"key_none":null,"key_enum":"value1","key_nested":{"key_str":"nested"},"arr_str":["val1"],"arr_int":[42],"arr_float":[3,1,4]', None),
        ('{"key_str": "val","key_int":42,"key_float":3.14,"key_bool":false,"key_none":null,"key_enum":"value1","key_nested":{"key_str":"nested"},"arr_str":["val1"],"arr_int":[42],"arr_float":[3.14,31.4]', None),
        ('{"key_str": "val","key_int":42,"key_float":3.14,"key_bool":false,"key_none":null,"key_enum":"value1","key_nested":{"key_str":"nested"},"arr_str":["val1"],"arr_int":[42],"arr_float":[3.14],"arr_bool":"', UnexpectedCharacterError),
        ('{"key_str": "val","key_int":42,"key_float":3.14,"key_bool":false,"key_none":null,"key_enum":"value1","key_nested":{"key_str":"nested"},"arr_str":["val1"],"arr_int":[42],"arr_float":[3.14],"arr_bool":t', UnexpectedCharacterError),
        ('{"key_str": "val","key_int":42,"key_float":3.14,"key_bool":false,"key_none":null,"key_enum":"value1","key_nested":{"key_str":"nested"},"arr_str":["val1"],"arr_int":[42],"arr_float":[3.14],"arr_bool":r', UnexpectedCharacterError),
        ('{"key_str": "val","key_int":42,"key_float":3.14,"key_bool":false,"key_none":null,"key_enum":"value1","key_nested":{"key_str":"nested"},"arr_str":["val1"],"arr_int":[42],"arr_float":[3.14],"arr_bool":[]', None),
        ('{"key_str": "val","key_int":42,"key_float":3.14,"key_bool":false,"key_none":null,"key_enum":"value1","key_nested":{"key_str":"nested"},"arr_str":["val1"],"arr_int":[42],"arr_float":[3.14],"arr_bool":[true,false,true]', None),
        ('{"key_str": "val","key_int":42,"key_float":3.14,"key_bool":false,"key_none":null,"key_enum":"value1","key_nested":{"key_str":"nested"},"arr_str":["val1"],"arr_int":[42],"arr_float":[3.14],"arr_bool":[true],"arr_none":n', UnexpectedCharacterError),
        ('{"key_str": "val","key_int":42,"key_float":3.14,"key_bool":false,"key_none":null,"key_enum":"value1","key_nested":{"key_str":"nested"},"arr_str":["val1"],"arr_int":[42],"arr_float":[3.14],"arr_bool":[true],"arr_none":f', UnexpectedCharacterError),
        ('{"key_str": "val","key_int":42,"key_float":3.14,"key_bool":false,"key_none":null,"key_enum":"value1","key_nested":{"key_str":"nested"},"arr_str":["val1"],"arr_int":[42],"arr_float":[3.14],"arr_bool":[true],"arr_none":[]', None),
        ('{"key_str": "val","key_int":42,"key_float":3.14,"key_bool":false,"key_none":null,"key_enum":"value1","key_nested":{"key_str":"nested"},"arr_str":["val1"],"arr_int":[42],"arr_float":[3.14],"arr_bool":[true],"arr_none":[null]', None),
        ('{"key_str": "val","key_int":42,"key_float":3.14,"key_bool":false,"key_none":null,"key_enum":"value1","key_nested":{"key_str":"nested"},"arr_str":["val1"],"arr_int":[42],"arr_float":[3.14],"arr_bool":[true],"arr_none":[null],"arr_enum"', None),
        ('{"key_str": "val","key_int":42,"key_float":3.14,"key_bool":false,"key_none":null,"key_enum":"value1","key_nested":{"key_str":"nested"},"arr_str":["val1"],"arr_int":[42],"arr_float":[3.14],"arr_bool":[true],"arr_none":[null],"arr_enum":[', None),
        ('{"key_str": "val","key_int":42,"key_float":3.14,"key_bool":false,"key_none":null,"key_enum":"value1","key_nested":{"key_str":"nested"},"arr_str":["val1"],"arr_int":[42],"arr_float":[3.14],"arr_bool":[true],"arr_none":[null],"arr_enum":["val', None),
        ('{"key_str": "val","key_int":42,"key_float":3.14,"key_bool":false,"key_none":null,"key_enum":"value1","key_nested":{"key_str":"nested"},"arr_str":["val1"],"arr_int":[42],"arr_float":[3.14],"arr_bool":[true],"arr_none":[null],"arr_enum":["foo', None),
        ('{"key_str": "val","key_int":42,"key_float":3.14,"key_bool":false,"key_none":null,"key_enum":"value1","key_nested":{"key_str":"nested"},"arr_str":["val1"],"arr_int":[42],"arr_float":[3.14],"arr_bool":[true],"arr_none":[null],"arr_enum":["value1"', None),
        ('{"key_str": "val","key_int":42,"key_float":3.14,"key_bool":false,"key_none":null,"key_enum":"value1","key_nested":{"key_str":"nested"},"arr_str":["val1"],"arr_int":[42],"arr_float":[3.14],"arr_bool":[true],"arr_none":[null],"arr_enum":["foobar"', ParsePrimitiveError),
        ('{"key_str": "val","key_int":42,"key_float":3.14,"key_bool":false,"key_none":null,"key_enum":"value1","key_nested":{"key_str":"nested"},"arr_str":["val1"],"arr_int":[42],"arr_float":[3.14],"arr_bool":[true],"arr_none":[null],"arr_enum":["value1","value2"]', None),
        ('{"key_str": "val","key_int":42,"key_float":3.14,"key_bool":false,"key_none":null,"key_enum":"value1","key_nested":{"key_str":"nested"},"arr_str":["val1"],"arr_int":[42],"arr_float":[3.14],"arr_bool":[true],"arr_none":[null],"arr_enum":["value1","value2"],"arr_nested":[]', None),
        ('{"key_str": "val","key_int":42,"key_float":3.14,"key_bool":false,"key_none":null,"key_enum":"value1","key_nested":{"key_str":"nested"},"arr_str":["val1"],"arr_int":[42],"arr_float":[3.14],"arr_bool":[true],"arr_none":[null],"arr_enum":["value1","value2"],"arr_nested":[3', UnexpectedCharacterError),
        ('{"key_str": "val","key_int":42,"key_float":3.14,"key_bool":false,"key_none":null,"key_enum":"value1","key_nested":{"key_str":"nested"},"arr_str":["val1"],"arr_int":[42],"arr_float":[3.14],"arr_bool":[true],"arr_none":[null],"arr_enum":["value1","value2"],"arr_nested":[p', UnexpectedCharacterError),
        ('{"key_str": "val","key_int":42,"key_float":3.14,"key_bool":false,"key_none":null,"key_enum":"value1","key_nested":{"key_str":"nested"},"arr_str":["val1"],"arr_int":[42],"arr_float":[3.14],"arr_bool":[true],"arr_none":[null],"arr_enum":["value1","value2"],"arr_nested":[{p', UnexpectedCharacterError),
        ('{"key_str": "val","key_int":42,"key_float":3.14,"key_bool":false,"key_none":null,"key_enum":"value1","key_nested":{"key_str":"nested"},"arr_str":["val1"],"arr_int":[42],"arr_float":[3.14],"arr_bool":[true],"arr_none":[null],"arr_enum":["value1","value2"],"arr_nested":[{"key_str":3', UnexpectedCharacterError),
        ('{"key_str": "val","key_int":42,"key_float":3.14,"key_bool":false,"key_none":null,"key_enum":"value1","key_nested":{"key_str":"nested"},"arr_str":["val1"],"arr_int":[42],"arr_float":[3.14],"arr_bool":[true],"arr_none":[null],"arr_enum":["value1","value2"],"arr_nested":[{"key_str":', None),
        ('{"key_str": "val","key_int":42,"key_float":3.14,"key_bool":false,"key_none":null,"key_enum":"value1","key_nested":{"key_str":"nested"},"arr_str":["val1"],"arr_int":[42],"arr_float":[3.14],"arr_bool":[true],"arr_none":[null],"arr_enum":["value1","value2"],"arr_nested":[{"key_str":"nested1"},{"key_str":"nested2"}]}', None),
        ('{"key_str": "val","key_int":42,"key_float":3.14,"key_bool":false,"key_none":null,"key_enum":"value1","key_nested":{"key_str":"nested"},"arr_str":["val1"],"arr_int":[42],"arr_float":[3.14],"arr_bool":[true],"arr_none":[null],"arr_enum":["value1","value2"],"arr_nested":[{"key_str":"nested1"},{"key_str":"nested2"}]}}', ObjectAlreadyClosedError),
    ],
)
# fmt: on
@pytest.mark.anyio
async def test_json_demux__parse_incorrect_stream__assert_error(
    stream: str, MaybeExpectedError: Type[Exception] | None
):
    class SObject(JMux):
        class SNested(JMux):
            key_str: AwaitableValue[str]
        
        class SEnum(Enum):
            VALUE1 = "value1"
            VALUE2 = "value2"

        key_str: AwaitableValue[str]
        key_int: AwaitableValue[int]
        key_float: AwaitableValue[float]
        key_bool: AwaitableValue[bool]
        key_none: AwaitableValue[NoneType]
        key_stream: StreamableValues[str]
        key_enum: AwaitableValue[SEnum]
        key_nested: AwaitableValue[SNested]

        arr_str: StreamableValues[str]
        arr_int: StreamableValues[int]
        arr_float: StreamableValues[float]
        arr_bool: StreamableValues[bool]
        arr_none: StreamableValues[NoneType]
        arr_enum: StreamableValues[SEnum]
        arr_nested: StreamableValues[SNested]

    s_object = SObject()

    if MaybeExpectedError:
        with pytest.raises(MaybeExpectedError):
            for ch in stream:
                await s_object.feed_char(ch)
    else:
        for ch in stream:
            await s_object.feed_char(ch)



# fmt: off
@pytest.mark.parametrize(
    "stream,MaybeExpectedError",
    [
        ("b", UnexpectedCharacterError),
        ("{", None),
        ("{p", UnexpectedCharacterError),
        ('{"', None),
        ('{""', EmptyKeyError),
        ('{"no_actual_key"', MissingAttributeError),
        ('{"key_str"', None),
        ('{"key_str": ""', None),
        ('{"key_str": n', None),
        ('{"key_str": t', UnexpectedCharacterError),
        ('{"key_str": null', None),
        ('{"key_str": "val","key_int":4p', UnexpectedCharacterError),
        ('{"key_str": "val","key_int":n', None),
        ('{"key_str": "val","key_int":r', UnexpectedCharacterError),
        ('{"key_str": "val","key_int":null', None),
        ('{"key_str": "val","key_int":null,', None),
        ('{"key_str": "val","key_int":420', None),
        ('{"key_str": "val","key_int":-420', None),
        ('{"key_str": "val","key_int":42,"key_float":0', None),
        ('{"key_str": "val","key_int":42,"key_float":n', None),
        ('{"key_str": "val","key_int":42,"key_float":l', UnexpectedCharacterError),
        ('{"key_str": "val","key_int":42,"key_float":null', None),
        ('{"key_str": "val","key_int":42,"key_float":null,', None),
        ('{"key_str": "val","key_int":42,"key_float":3.14,"key_bool":t', None),
        ('{"key_str": "val","key_int":42,"key_float":3.14,"key_bool":r', UnexpectedCharacterError),
        ('{"key_str": "val","key_int":42,"key_float":3.14,"key_bool":true,', None),
        ('{"key_str": "val","key_int":42,"key_float":3.14,"key_bool":n', None),
        ('{"key_str": "val","key_int":42,"key_float":3.14,"key_bool":null', None),
        ('{"key_str": "val","key_int":42,"key_float":3.14,"key_bool":null,', None),
        ('{"key_str": "val","key_int":42,"key_float":3.14,"key_bool":false,"key_enum":t', UnexpectedCharacterError),
        ('{"key_str": "val","key_int":42,"key_float":3.14,"key_bool":false,"key_enum":"', None),
        ('{"key_str": "val","key_int":42,"key_float":3.14,"key_bool":false,"key_enum":n', None),
        ('{"key_str": "val","key_int":42,"key_float":3.14,"key_bool":false,"key_enum":null,', None),
        ('{"key_str": "val","key_int":42,"key_float":3.14,"key_bool":false,"key_enum":"value1"', None),
        ('{"key_str": "val","key_int":42,"key_float":3.14,"key_bool":false,"key_enum":"value1","key_nested":{', None),
        ('{"key_str": "val","key_int":42,"key_float":3.14,"key_bool":false,"key_enum":"value1","key_nested":n', None),
        ('{"key_str": "val","key_int":42,"key_float":3.14,"key_bool":false,"key_enum":"value1","key_nested":k', UnexpectedCharacterError),
        ('{"key_str": "val","key_int":42,"key_float":3.14,"key_bool":false,"key_enum":"value1","key_nested":null', None),
        ('{"key_str": "val","key_int":42,"key_float":3.14,"key_bool":false,"key_enum":"value1","key_nested":null,', None),
        ('{"key_str": "val","key_int":42,"key_float":3.14,"key_bool":false,"key_enum":"value1","key_nested":{p', UnexpectedCharacterError),
        ('{"key_str": "val","key_int":42,"key_float":3.14,"key_bool":false,"key_enum":"value1","key_nested":{n', UnexpectedCharacterError),
        ('{"key_str": "val","key_int":42,"key_float":3.14,"key_bool":false,"key_enum":"value1","key_nested":{"key_str":"nested"},', None),
    ],
)
# fmt: on
@pytest.mark.anyio
async def test_json_demux__parse_incorrect_stream_with_optionals__assert_error(
    stream: str, MaybeExpectedError: Type[Exception] | None
):
    class SObject(JMux):
        class SNested(JMux):
            key_str: AwaitableValue[str]

        class SEnum(Enum):
            VALUE1 = "value1"
            VALUE2 = "value2"

        key_str: AwaitableValue[str | NoneType]
        key_int: AwaitableValue[int | NoneType]
        key_float: AwaitableValue[float | NoneType]
        key_bool: AwaitableValue[bool | NoneType]
        key_enum: AwaitableValue[SEnum | NoneType]
        key_nested: AwaitableValue[SNested | NoneType]

    s_object = SObject()

    if MaybeExpectedError:
        with pytest.raises(MaybeExpectedError):
            for ch in stream:
                await s_object.feed_char(ch)
    else:
        for ch in stream:
            await s_object.feed_char(ch)



from enum import Enum

import pytest
from pydantic import BaseModel

from jmux.awaitable import AwaitableValue, StreamableValues
from jmux.demux import JMux
from jmux.error import ObjectMissmatchedError


class SEnum(Enum):
    VALUE1 = "value1"
    VALUE2 = "value2"


class CorrectJMux_1(JMux):
    class NestedJMux(JMux):
        nested_key: AwaitableValue[str]

    key_str: AwaitableValue[str]
    key_int: AwaitableValue[int]
    key_float: AwaitableValue[float]
    key_bool: AwaitableValue[bool]
    key_none: AwaitableValue[None]
    key_enum: AwaitableValue[SEnum]
    key_nested: AwaitableValue[NestedJMux]

    key_stream: StreamableValues[str]

    arr_str: StreamableValues[str]
    arr_int: StreamableValues[int]
    arr_float: StreamableValues[float]
    arr_bool: StreamableValues[bool]
    arr_none: StreamableValues[None]
    arr_enum: StreamableValues[SEnum]
    arr_nested: StreamableValues[NestedJMux]


class CorrectPydantic_1(BaseModel):
    class NestedPydantic(BaseModel):
        nested_key: str

    key_str: str
    key_int: int
    key_float: float
    key_bool: bool
    key_none: None
    key_stream: str
    key_enum: SEnum
    key_nested: NestedPydantic

    arr_str: list[str]
    arr_int: list[int]
    arr_float: list[float]
    arr_bool: list[bool]
    arr_none: list[None]
    arr_enum: list[SEnum]
    arr_nested: list[NestedPydantic]


class CorrectJMux_2(JMux):
    key_str: AwaitableValue[str | None]
    key_int: AwaitableValue[int | None]
    key_float: AwaitableValue[float | None]
    key_bool: AwaitableValue[bool | None]


class CorrectPydantic_2(BaseModel):
    key_str: str | None
    key_int: int | None
    key_float: float | None
    key_bool: bool | None


class IncorrectJMux_1(JMux):
    key_str: AwaitableValue[str]


class IncorrectPydantic_1(BaseModel):
    key_str: str | None


class IncorrectJMux_2(JMux):
    key_str: AwaitableValue[str | None]


class IncorrectPydantic_2(BaseModel):
    key_str: str


class IncorrectJMux_3(JMux):
    class NestedJMux(JMux):
        nested_key: AwaitableValue[str | None]

    key_nested: AwaitableValue[NestedJMux]


class IncorrectPydantic_3(BaseModel):
    class NestedPydantic(BaseModel):
        nested_key: str

    key_nested: NestedPydantic


@pytest.mark.parametrize(
    "TargetJMux,TargetPydantic,MaybeExpectedError",
    [
        (CorrectJMux_1, CorrectPydantic_1, None),
        (CorrectJMux_2, CorrectPydantic_2, None),
        (IncorrectJMux_1, IncorrectPydantic_1, ObjectMissmatchedError),
        (IncorrectJMux_2, IncorrectPydantic_2, ObjectMissmatchedError),
        (IncorrectJMux_3, IncorrectPydantic_3, ObjectMissmatchedError),
    ],
)
@pytest.mark.anyio
async def test_json_demux__validate_pydantic(
    TargetJMux: type[JMux],
    TargetPydantic: type[BaseModel],
    MaybeExpectedError: type[Exception] | None,
):
    if MaybeExpectedError:
        with pytest.raises(MaybeExpectedError):
            TargetJMux.assert_conforms_to(TargetPydantic)
    else:
        TargetJMux.assert_conforms_to(TargetPydantic)



from asyncio import Event, Queue
from enum import Enum
from types import NoneType
from typing import (
    AsyncGenerator,
    Protocol,
    Set,
    Type,
    cast,
    runtime_checkable,
)

from jmux.error import NothingEmittedError, SinkClosedError
from jmux.helpers import extract_types_from_generic_alias


class SinkType(Enum):
    STREAMABLE_VALUES = "StreamableValues"
    AWAITABLE_VALUE = "AwaitableValue"


class UnderlyingGenericMixin[T]:
    """
    A mixin class that provides methods for inspecting the generic types of a 
    class at runtime.
    """

    def get_underlying_generics(self) -> Set[Type[T]]:
        """
        Returns the underlying generic types of the class.

        Raises:
            TypeError: If the class is not initialized with a defined generic type.

        Returns:
            A set of the underlying generic types.
        """
        # `__orig_class__` is only set after the `__init__` method is called
        if not hasattr(self, "__orig_class__"):
            raise TypeError(
                "AwaitableValue must be initialized with a defined generic type."
            )

        Origin = getattr(self, "__orig_class__")
        _, type_set = extract_types_from_generic_alias(Origin)
        return type_set

    def get_underlying_main_generic(self) -> Type[T]:
        """
        Returns the main underlying generic type of the class.
        This is the generic type that is not NoneType.

        Returns:
            The main underlying generic type.
        """
        underlying_generics = self.get_underlying_generics()
        if len(underlying_generics) == 1:
            return underlying_generics.pop()
        remaining = {g for g in underlying_generics if g is not NoneType}
        return remaining.pop()


@runtime_checkable
class IAsyncSink[T](Protocol):
    """
    An asynchronous sink protocol that defines a common interface for putting, closing,
    and retrieving values from a sink.
    """

    def get_underlying_generics(self) -> Set[Type[T]]:
        """Return the underlying generic type of the sink."""
        ...

    def get_underlying_main_generic(self) -> Type[T]:
        """Return the underlying non-NoneType generic type of the sink."""
        ...

    async def put(self, item: T):
        """Put an item into the sink."""
        ...

    async def close(self):
        """Close the sink."""
        ...

    async def ensure_closed(self):
        """Ensure the sink is closed."""
        ...

    def get_current(self) -> T:
        """Get the current value from the sink."""
        ...

    def get_sink_type(self) -> SinkType:
        """Get the type of the sink."""
        ...


class StreamableValues[T](UnderlyingGenericMixin[T]):
    """
    A class that represents a stream of values that can be asynchronously iterated over.
    It uses an asyncio.Queue to store the items and allows for putting items into the
    stream and closing it when no more items will be added.
    """

    def __init__(self):
        self._queue = Queue[T | None]()
        self._last_item: T | None = None
        self._closed = False

    def get_underlying_generics(self) -> Set[Type[T]]:
        """
        Returns the underlying generic types of the class.

        Raises:
            TypeError: If the class does not have exactly one underlying type.

        Returns:
            A set of the underlying generic types.
        """
        generic = super().get_underlying_generics()
        if len(generic) != 1:
            raise TypeError("StreamableValues must have exactly one underlying type.")
        return generic

    async def put(self, item: T):
        """
        Puts an item into the stream.

        Args:
            item: The item to put into the stream.

        Raises:
            ValueError: If the stream is closed.
        """
        if self._closed:
            raise ValueError("Cannot put item into a closed sink.")
        self._last_item = item
        await self._queue.put(item)

    async def close(self):
        """
        Closes the stream.

        Raises:
            SinkClosedError: If the stream is already closed.
        """
        if self._closed:
            raise SinkClosedError(
                f"SinkType {self.get_sink_type()}[{self.get_underlying_main_generic()}]"
                  + " is already closed."
            )
        self._closed = True
        await self._queue.put(None)

    async def ensure_closed(self):
        """
        Ensures that the stream is closed.
        If the stream is already closed, this method does nothing.
        """
        if self._closed:
            return
        await self.close()

    def get_current(self) -> T:
        """
        Returns the last item that was put into the stream.

        Raises:
            ValueError: If no items have been put into the stream yet.

        Returns:
            The last item that was put into the stream.
        """
        if self._last_item is None:
            raise ValueError("StreamableValues has not received any items yet.")
        return self._last_item

    def get_sink_type(self) -> SinkType:
        """
        Returns the type of the sink.

        Returns:
            The type of the sink.
        """
        return SinkType.STREAMABLE_VALUES

    def __aiter__(self):
        return self._stream()

    async def _stream(self) -> AsyncGenerator[T, None]:
        while True:
            item = await self._queue.get()
            if item is None and self._closed:
                break
            if item is None:
                raise ValueError("Received None item, but the sink is not closed.")
            yield item


class AwaitableValue[T](UnderlyingGenericMixin[T]):
    """
    A class that represents a value that will be available in the future.
    It can be awaited to get the value, and it can only be set once.
    """

    def __init__(self):
        self._is_closed = False
        self._event = Event()
        self._value: T | None = None

    async def put(self, value: T):
        """
        Sets the value of the AwaitableValue.

        Args:
            value: The value to set.

        Raises:
            ValueError: If the value has already been set.
        """
        if self._value is not None or self._is_closed or self._event.is_set():
            raise ValueError("AwaitableValue can only be set once.")
        self._value = value
        self._event.set()

    async def close(self):
        """
        Closes the AwaitableValue.

        Raises:
            SinkClosedError: If the AwaitableValue is already closed.
            NothingEmittedError: If the AwaitableValue is closed without a value,
                and the underlying type is not NoneType.
        """
        if self._is_closed:
            raise SinkClosedError(
                f"SinkType {self.get_sink_type()}"
                +"[{self.get_underlying_main_generic().__name__}] is already closed."
            )
        elif not self._event.is_set() and NoneType in self.get_underlying_generics():
            self._event.set()
        elif not self._event.is_set():
            raise NothingEmittedError(
                "Trying to close non-NoneType AwaitableValue without a value."
            )
        self._is_closed = True

    async def ensure_closed(self):
        """
        Ensures that the AwaitableValue is closed.
        If the AwaitableValue is already closed, this method does nothing.
        """
        if self._is_closed:
            return
        await self.close()

    def get_current(self) -> T:
        """
        Returns the value of the AwaitableValue.

        Raises:
            ValueError: If the value has not been set yet.

        Returns:
            The value of the AwaitableValue.
        """
        if self._value is None:
            raise ValueError("AwaitableValue has not been set yet.")
        return self._value

    def get_sink_type(self) -> SinkType:
        """
        Returns the type of the sink.

        Returns:
            The type of the sink.
        """
        return SinkType.AWAITABLE_VALUE

    def __await__(self):
        return self._wait().__await__()

    async def _wait(self) -> T:
        await self._event.wait()
        if self._value is None and not self._event.is_set():
            raise ValueError("No value has been put into the sink.")
        return cast(T, self._value)



from enum import Enum
from typing import Set


class State(Enum):
    START = "start"
    END = "end"
    ERROR = "error"
    # expect
    EXPECT_KEY = "expect_key"
    EXPECT_COLON = "expect_colon"
    EXPECT_VALUE = "expect_value"
    EXPECT_COMMA_OR_EOC = "expect_comma_or_eoc"
    # parsing
    PARSING_KEY = "parsing_key"
    PARSING_STRING = "parsing_string"
    PARSING_INTEGER = "parsing_integer"
    PARSING_FLOAT = "parsing_float"
    PARSING_BOOLEAN = "parsing_boolean"
    PARSING_NULL = "parsing_null"
    PARSING_OBJECT = "parsing_object"


PRIMITIVE_STATES: Set[State] = {
    State.PARSING_INTEGER,
    State.PARSING_FLOAT,
    State.PARSING_BOOLEAN,
    State.PARSING_NULL,
}


class Mode(Enum):
    ROOT = "$"
    OBJECT = "object"
    ARRAY = "array"


OBJECT_OPEN = set("{")
OBJECT_CLOSE = set("}")
COLON = set(":")
ARRAY_OPEN = set("[")
ARRAY_CLOSE = set("]")
COMMA = set(",")
QUOTE = set('"')

NUMBER_OPEN = set("0123456789-")
BOOLEAN_OPEN = set("tf")
NULL_OPEN = set("n")

INTERGER_ALLOWED = set("0123456789")
FLOAT_ALLOWED = set("0123456789-+eE.")
BOOLEAN_ALLOWED = set("truefals")
NULL_ALLOWED = set("nul")

JSON_FALSE = "false"
JSON_TRUE = "true"
JSON_NULL = "null"
